-- Character Manager 2 by j2b2

---@alias CharacterManager2.ButtonState integer

---@class CharacterManager2.Character
---@field EquipSkin fun(player: player_t, oldSkin: string, newSkin: string)
---@field UnEquipSkin fun(player: player_t, oldSkin: string, newSkin: string)
---@field Spawn fun(player: player_t)
---@field Think fun(player: player_t)
---@field Land fun(player: player_t)
---@field CarryStart fun(player: player_t, carry_type: integer)
---@field CarryEnd fun(player: player_t, carry_type: integer)
---@field TakeDamage fun(player: player_t, playerMo: mobj_t, inflictor: mobj_t, source: mobj_t, damage: integer, damagetype: damage_type): override: boolean?
---@field CanDamage fun(player: player_t, mobj: mobj_t): override: boolean?
---@field JumpButton fun(player: player_t, button_state: CharacterManager2.ButtonState): override: boolean?
---@field Custom1 fun(player: player_t, button_state: CharacterManager2.ButtonState)
---@field Custom2 fun(player: player_t, button_state: CharacterManager2.ButtonState)
---@field Custom3 fun(player: player_t, button_state: CharacterManager2.ButtonState)
---@field TossFlag fun(player: player_t, button_state: CharacterManager2.ButtonState)
---@field BotTiccmd fun(player: player_t, cmd: ticcmd_t)

-- Cache global constants as accessing global variables is very slow

local L_BT_WEAPONMASK = BT_WEAPONMASK
local L_BT_WEAPONNEXT = BT_WEAPONNEXT
local L_BT_WEAPONPREV = BT_WEAPONPREV
local L_BT_ATTACK = BT_ATTACK
local L_BT_SPIN = BT_SPIN
local L_BT_TOSSFLAG = BT_TOSSFLAG
local L_BT_JUMP = BT_JUMP
local L_BT_FIRENORMAL = BT_FIRENORMAL
local L_BT_CUSTOM1 = BT_CUSTOM1
local L_BT_CUSTOM2 = BT_CUSTOM2
local L_BT_CUSTOM3 = BT_CUSTOM3

local L_MFE_JUSTHITFLOOR = MFE_JUSTHITFLOOR

local L_PF_SPINDOWN = PF_SPINDOWN
local L_PF_JUMPDOWN = PF_JUMPDOWN

local l_pw_shield = pw_shield
local l_pw_carry = pw_carry
local l_pw_super = pw_super

local L_CR_NONE = CR_NONE

local L_P_IsObjectOnGround = P_IsObjectOnGround

local function cm_print(...)
	local args = {...}
	local str_args = {}
	for _, v in ipairs(args) do
		table.insert(str_args, tostring(v))
	end
	print("CharacterManager2: "..table.concat(str_args, "    "))
end

local CM = {
	VERSION_MAJOR = 0,
	VERSION_MINOR = 4,
	STATE_PRESSED = 0, ---@type CharacterManager2.ButtonState
	STATE_HELD = 1, ---@type CharacterManager2.ButtonState
	STATE_RELEASED = 2, ---@type CharacterManager2.ButtonState
	skins = {},
	functions = {},
	events = {
		Think = {"PlayerThink"},
		EquipSkin = {"PlayerThink"},
		UnequipSkin = {"PlayerThink"},
		LineCollide = {"MobjLineCollide"},
		MoveBlocked = {"MobjMoveBlocked"},
		CanDamage = {"PlayerCanDamage"},
		Spawn = {"PlayerSpawn"},
		TakeDamage = {"MobjDamage"},
		Death = {"MobjDeath"},
		Fuse = {"MobjFuse"},
		ShouldTakeDamage = {"ShouldDamage"},
		ShieldSpawn = {"ShieldSpawn"},
		UseShield = {"ShieldSpecial"},
		TouchSpecial = {"TouchSpecial"},
		Land = {"PlayerThink"},
		SuperStart = {"PlayerThink"},
		SuperEnd = {"PlayerThink"},
		CarryStart = {"PlayerThink"},
		CarryEnd = {"PlayerThink"},
		ButtonsChanged = {"PlayerThink"},
		WeaponSlot = {"PlayerThink"},
		WeaponNext = {"PlayerThink"},
		WeaponPrev = {"PlayerThink"},
		Fire = {"PlayerThink"},
		TossFlag = {"PlayerThink"},
		FireNormal = {"PlayerThink"},
		Custom1 = {"PlayerThink"},
		Custom2 = {"PlayerThink"},
		Custom3 = {"PlayerThink"},
		JumpButton = {"JumpSpecial", "PlayerThink"},
		SpinButton = {"SpinSpecial", "PlayerThink"},
		AirSpin = {"JumpSpinSpecial", "PlayerThink"},
		JumpAbility = {"AbilitySpecial"},
		BotTiccmd = {"BotTiccmd"},
		BotRespawn = {"BotRespawn"},
		-- TODO: Figure out what to do about HUDDraw event
	},
	loadedHooks = {},
}
rawset(_G, "CharacterManager2", CM)

--TODO: Handle using newer version

--[[
	TODO: Dedcide wether AirSpin/GroundSpin should send release while jumping and wether AirSpin/GroundSpin should exist at all.
	TODO: Add Events:
	* Inflict Damage (Maybe)
	* Get/Lose shield (Maybe)
	* Enter/Exit Water
	* Flip Gravity
	* Finish level
]]

local function GetSkinFunction(player, functionName)
	local l_realmo = player.realmo
	if not (l_realmo and l_realmo.valid) then return end
	local skinTable = CharacterManager2.skins[l_realmo.skin]
	if skinTable then
		return skinTable[functionName]
	end
end

function CharacterManager2.AddHook(hook)
	cm_print('Adding hook "'..hook..'"')
	if CharacterManager2.functions[hook] == nil then
		cm_print("Missing function "..hook)
			return false
	end
	addHook(hook, function(...) return CharacterManager2.functions[hook](...) end, MT_PLAYER)
	CharacterManager2.loadedHooks[hook] = true
end

function CharacterManager2.Register(skinTbl)
	local cm = CharacterManager2
	for event, hooks in pairs(cm.events) do
		if type(skinTbl[event]) == "function" then
			for _, hook in ipairs(hooks) do
				if not cm.loadedHooks[hook] then
					cm.AddHook(hook)
				end
			end
		end
	end
	if cm.skins[skinTbl.name] == nil then
		cm.skins[skinTbl.name] = skinTbl
	else
		error("Tried to register a skin twice.")
	end
end

function CharacterManager2.functions.AbilitySpecial(player)
	local func = GetSkinFunction(player, "JumpAbility")
	if func then return func(player) end
end

function CharacterManager2.functions.BotRespawn(playerMobj, botMobj, ...)
	if botMobj == nil or botMobj.valid == false then return end
	local botPlayer = botMobj.player
	if botPlayer == nil or botPlayer.valid == false then return end
	local func = GetSkinFunction(botPlayer, "BotRespawn")
	if func then return func(botPlayer, botMobj, playerMobj, ...) end
end

function CharacterManager2.functions.BotTiccmd(player, ...)
	local func = GetSkinFunction(player, "BotTiccmd")
	if func then return func(player, ...) end
end

function CharacterManager2.functions.JumpSpecial(player)
	local func = GetSkinFunction(player, "JumpButton")
	if func then
		local state
		if player.pflags & L_PF_JUMPDOWN then
			state = CharacterManager2.STATE_HELD
		else
			state = CharacterManager2.STATE_PRESSED
		end
		return func(player, state)
	end
end

function CharacterManager2.functions.JumpSpinSpecial(player)
	local func = GetSkinFunction(player, "AirSpin")
	if func then
		local state
		if player.pflags & L_PF_SPINDOWN then
			state = CharacterManager2.STATE_HELD
		else
			state = CharacterManager2.STATE_PRESSED
		end
		player.CM_AirSpin = true
		return func(player, state)
	end
end

function CharacterManager2.functions.MobjDamage(playerMo, ...)
	local player = playerMo.player
	if player == nil or player.valid == false then return end
	local func = GetSkinFunction(player, "TakeDamage")
	if func then return func(player, playerMo, ...) end
end

function CharacterManager2.functions.MobjDeath(playerMo, ...)
	local player = playerMo.player
	local func = GetSkinFunction(player, "Death")
	if func then return func(player, playerMo, ...) end
end

function CharacterManager2.functions.MobjFuse(playerMo, ...)
	local player = playerMo.player
	if player == nil or player.valid == false then return end
	local func = GetSkinFunction(player, "Fuse")
	if func then return func(player, playerMo, ...) end
end

function CharacterManager2.functions.MobjLineCollide(playerMo, ...)
	local player = playerMo.player
	if player == nil or player.valid == false then return end
	local func = GetSkinFunction(player, "LineCollide")
	if func then return func(player, playerMo, ...) end
end

function CharacterManager2.functions.MobjMoveBlocked(playerMo, ...)
	local player = playerMo.player
	if player == nil or player.valid == false then return end
	local func = GetSkinFunction(player, "MoveBlocked")
	if func then return func(player, playerMo, ...) end
end

function CharacterManager2.functions.PlayerCanDamage(player, ...)
	local func = GetSkinFunction(player, "CanDamage")
	if func then return func(player, ...) end
end

function CharacterManager2.functions.PlayerSpawn(player, ...)
	local func = GetSkinFunction(player, "Spawn")
	if func then return func(player, ...) end
end

function CharacterManager2.functions.PlayerThink(player)

	local playerRealMo = player.realmo
	if playerRealMo == nil or not playerRealMo.valid then return end

	local oldSkin = player.CM_OldSkin
	local newSkin = playerRealMo.skin
	local skinTable = CharacterManager2.skins[newSkin]

	if oldSkin ~= newSkin then
		local oldSkinTable = CharacterManager2.skins[oldSkin]

		if oldSkinTable ~= nil then
			local func = oldSkinTable.UnEquipSkin
			if func then func(player, oldSkin, newSkin) end
		end

		if skinTable == nil then
			player.CM_OldButtons = nil
			player.CM_OldCarry = nil
			player.CM_OldSuper = nil
			player.CM_AirSpin = nil
		else
			local func = skinTable.EquipSkin
			if func then func(player, oldSkin, newSkin) end
		end

		player.CM_OldSkin = newSkin
	end

	if skinTable == nil then return end

	if player.CM_OldButtons == nil then player.CM_OldButtons = player.cmd.buttons end
	if player.CM_OldCarry == nil then player.CM_OldCarry = player.powers[l_pw_carry] end
	if player.CM_OldSuper == nil then player.CM_OldSuper = player.powers[l_pw_super] > 0 end

	local playerMo = player.mo
	local playerMoValid
	if playerMo then playerMoValid = playerMo.valid end

	local oldButtons = player.CM_OldButtons
	local newButtons = player.cmd.buttons

	if playerMoValid and (playerMo.eflags & L_MFE_JUSTHITFLOOR) then
		local func = skinTable.Land
		if func then func(player) end
	end

	local carry = player.powers[l_pw_carry]
	local oldCarry = player.CM_OldCarry
	if carry ~= oldCarry then
		if oldCarry ~= L_CR_NONE then
			local func = skinTable.CarryEnd
			if func then func(player, oldCarry) end
		end
		if carry ~= L_CR_NONE then
			local func = skinTable.CarryStart
			if func then func(player, carry) end
		end
	end

	local isSuper = player.powers[l_pw_super] > 0
	if isSuper ~= player.CM_OldSuper then
		local func
		if isSuper then
			func = skinTable.SuperStart
		else
			func = skinTable.SuperEnd
		end
		if func then func(player) end
	end

	local func = skinTable.WeaponNext
	if func then
		local oldState = oldButtons & L_BT_WEAPONNEXT
		local newState = newButtons & L_BT_WEAPONNEXT
		if oldState and newState then
			func(player, CharacterManager2.STATE_HELD)
		elseif oldState < newState then
			func(player, CharacterManager2.STATE_PRESSED)
		elseif oldState > newState then
			func(player, CharacterManager2.STATE_RELEASED)
		end
	end

	local func = skinTable.WeaponPrev
	if func then
		local oldState = oldButtons & L_BT_WEAPONPREV
		local newState = newButtons & L_BT_WEAPONPREV
		if oldState and newState then
			func(player, CharacterManager2.STATE_HELD)
		elseif oldState < newState then
			func(player, CharacterManager2.STATE_PRESSED)
		elseif oldState > newState then
			func(player, CharacterManager2.STATE_RELEASED)
		end
	end

	local func = skinTable.Fire
	if func then
		local oldState = oldButtons & L_BT_ATTACK
		local newState = newButtons & L_BT_ATTACK
		if oldState and newState then
			func(player, CharacterManager2.STATE_HELD)
		elseif oldState < newState then
			func(player, CharacterManager2.STATE_PRESSED)
		elseif oldState > newState then
			func(player, CharacterManager2.STATE_RELEASED)
		end
	end

	local func = skinTable.TossFlag
	if func then
		local oldState = oldButtons & L_BT_TOSSFLAG
		local newState = newButtons & L_BT_TOSSFLAG
		if oldState and newState then
			func(player, CharacterManager2.STATE_HELD)
		elseif oldState < newState then
			func(player, CharacterManager2.STATE_PRESSED)
		elseif oldState > newState then
			func(player, CharacterManager2.STATE_RELEASED)
		end
	end

	local func = skinTable.FireNormal
	if func then
		local oldState = oldButtons & L_BT_FIRENORMAL
		local newState = newButtons & L_BT_FIRENORMAL
		if oldState and newState then
			func(player, CharacterManager2.STATE_HELD)
		elseif oldState < newState then
			func(player, CharacterManager2.STATE_PRESSED)
		elseif oldState > newState then
			func(player, CharacterManager2.STATE_RELEASED)
		end
	end

	local func = skinTable.Custom1
	if func then
		local oldState = oldButtons & L_BT_CUSTOM1
		local newState = newButtons & L_BT_CUSTOM1
		if oldState and newState then
			func(player, CharacterManager2.STATE_HELD)
		elseif oldState < newState then
			func(player, CharacterManager2.STATE_PRESSED)
		elseif oldState > newState then
			func(player, CharacterManager2.STATE_RELEASED)
		end
	end

	local func = skinTable.Custom2
	if func then
		local oldState = oldButtons & L_BT_CUSTOM2
		local newState = newButtons & L_BT_CUSTOM2
		if oldState and newState then
			func(player, CharacterManager2.STATE_HELD)
		elseif oldState < newState then
			func(player, CharacterManager2.STATE_PRESSED)
		elseif oldState > newState then
			func(player, CharacterManager2.STATE_RELEASED)
		end
	end

	local func = skinTable.Custom3
	if func then
		local oldState = oldButtons & L_BT_CUSTOM3
		local newState = newButtons & L_BT_CUSTOM3
		if oldState and newState then
			func(player, CharacterManager2.STATE_HELD)
		elseif oldState < newState then
			func(player, CharacterManager2.STATE_PRESSED)
		elseif oldState > newState then
			func(player, CharacterManager2.STATE_RELEASED)
		end
	end

	if oldButtons ~= newButtons then

		if (oldButtons & L_BT_SPIN) > (newButtons & L_BT_SPIN) then
			if playerMoValid and not L_P_IsObjectOnGround(playerMo) then
				local func = skinTable.AirSpin
				if func then func(player, CharacterManager2.STATE_RELEASED) end
			end
			local func = skinTable.SpinButton
			if func then func(player, CharacterManager2.STATE_RELEASED) end
		end

		if (oldButtons & L_BT_JUMP) > (newButtons & L_BT_JUMP) then
			local func = skinTable.JumpButton
			if func then func(player, CharacterManager2.STATE_RELEASED) end
		end

		local func = skinTable.WeaponSlot
		if func then
			local slot = newButtons & L_BT_WEAPONMASK
			if slot then func(player, slot) end
		end

		local func = skinTable.ButtonsChanged
		if func then func(player, oldButtons, newButtons) end
	end

	local func = skinTable.Think
	if func then func(player) end

	player.CM_OldButtons = newButtons
	player.CM_OldCarry = carry
	player.CM_OldSuper = isSuper

end

function CharacterManager2.functions.ShieldSpawn(player, ...)
	local func = GetSkinFunction(player, "ShieldSpawn")
	if func then return func(player, ...) end
end

function CharacterManager2.functions.ShieldSpecial(player)
	local func = GetSkinFunction(player, "UseShield")
	if func then return func(player) end
end

function CharacterManager2.functions.SpinSpecial(player)
	local func = GetSkinFunction(player, "SpinButton")
	if func then
		local state
		if player.pflags & L_PF_SPINDOWN then
			state = CharacterManager2.STATE_HELD
		else
			state = CharacterManager2.STATE_PRESSED
		end
		return func(player, state)
	end
end

function CharacterManager2.functions.ShouldDamage(playerMo, ...)
	local player = playerMo.player
	if player == nil or player.valid == false then return end
	local func = GetSkinFunction(player, "ShouldTakeDamage")
	if func then return func(player, playerMo, ...) end
end

function CharacterManager2.functions.TouchSpecial()
end
